/*
 * treept
 *
 * Copyright (C) 1996 Pavel Zaboj (Pavel.Zaboj@vse.cz)
 * Plan 9 port C) 2008 Peter A. Cejchan (cej@gli.cas.cz)
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2, or (at your option) any
 * later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define  VERSION	"980310"

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <ctype.h>

#ifdef _PLAN9
#define CNS 1000
#define CNE 1000
#endif

typedef struct _node
  {
    unsigned int B, M, E;
    unsigned int th;
    struct _node *next, *prev;
  }
node;

typedef struct _nod
  {
    unsigned B, M, E;
  }
nod;


typedef struct _vals
  {
    double val;
    struct _vals *next;
  }
vals;

typedef struct _stack
  {
    int s;
    struct _stack *next;
    struct _stack *prev;
  }
stack;

typedef struct _dvo
  {
    float val;
    unsigned group;
    struct _dvo *next;
  }
dvo;

typedef struct _cvo
  {
    float v;
    float v1;
    struct _cvo *next;
  }
cvo;

node *s1 = NULL;

void
set_node (char c, int p, int n)
{
  node *s2;

  if (c == 'B')
    {
      s2 = (node *) malloc (sizeof (node));
      s2->th = p;
      s2->B = n;
      s2->M = s2->E = 0;
      s2->next = NULL;
      s2->prev = NULL;
      if (s1 != NULL)
	{
	  while (s1->next != NULL)
	    s1 = s1->next;
	  s1->next = s2;
	  s2->prev = s1;
	}
      s1 = s2;
    }
  else
    {
      if (s1->th != p)
	{
	  if (s1->th > p)
	    while (s1->th > p)
	      s1 = s1->prev;
	  else
	    while (s1->th < p)
	      s1 = s1->next;
	}
      if (c == 'M')
	s1->M = n;
      else
	s1->E = n;
    }

}

stack *sp = NULL;

void
push (int n)
{
  stack *new;
  if (sp == NULL)
    {
      sp = malloc (sizeof (stack));
      sp->next = sp->prev = NULL;
      sp->s = n;
      return;
    }
  new = malloc (sizeof (stack));
  new->prev = sp;
  new->next = NULL;
  new->s = n;
  sp->next = new;
  sp = new;
}

int
pull (void)
{
  stack *new;
  int a;
  if (sp == NULL)
    return (-1);
  a = sp->s;
  new = sp->prev;
  free (sp);
  sp = new;
  return (a);
}

FILE *foutput;
double *mat, *ends, min, sx, sy, vx, vy;
unsigned NP, NN, NE, NS, *perm, *R, pmin, ps, peq, pmax;
unsigned GR, CP, EX;		/* Group to cut, Cut percent, Extreme, to Group */
int TO;
nod *nodes;
dvo *ds;
cvo *cs;
char verbose, Verbose, mt, xfl, pfl;
unsigned NPERM = 1;

void
test (void)
{
#ifdef _PLAN9
  unsigned k, gn[CNS + 1], i, j, n;
  double S, gval[CNS + 1], gval2[CNS + 1], val;
#else
  unsigned k, gn[NS + 1], i, j, n;
  double S, gval[NS + 1], gval2[NS + 1], val;
#endif

  static unsigned prg = 0;
  static unsigned pst = 0;

#ifdef _PLAN9
  double Val[CNE + 1];
  unsigned Group[CNE + 1];
  unsigned Extr[CNE + 1];
#else
  double Val[NE + 1];
  unsigned Group[NE + 1];
  unsigned Extr[NE + 1];
#endif



  unsigned pocet, tr, notr;

  if (pfl && !verbose && !Verbose && prg == pst)
    {
      printf ("%3u%%\n", (unsigned) ((double) prg / NPERM * 100.0));
      pst += (unsigned) ceil ((double) NPERM / 10.0);
    }
  prg++;
  S = 0;
  if (xfl == 0)
    {
      for (k = 0; k < NN; k++)
	{
	  val = 0;
	  n = 0;
	  for (i = nodes[k].B; i <= nodes[k].M - 1; i++)
	    for (j = nodes[k].M; j <= nodes[k].E; j++)
	      {
		if (mt == 0)
		  val = val + fabs (ends[perm[i]] - ends[perm[j]]);
		else
		  val = val + mat[perm[i] * NE + perm[j]];
		n++;
	      }
	  val = val / n;
	  S = S + val;
	}
    }
  else
    {
      if (xfl == 1 || xfl == 2)	/* variance test */
	{
	  for (i = 0; i <= NS; i++)
	    gval2[i] = gval[i] = gn[i] = 0;

	  pocet = 0;
	  for (j = 0; j < NE; j++)
	    {
	      Val[j] = ds[j].val;
	      k = Group[j] = ds[perm[j]].group;
	      gn[k]++;
	      if (Group[j] == GR)
		pocet++;
	    }
/*
	  if (verbose)
	    {
	      for (i = 0; i <= NS; i++)
		printf ("%d ", gn[i]);
	      printf ("\n");
	    }
*/
	  if (verbose && CP != 100 && ps == 1)
	    printf ("%d elements in group %d, truncated %d%%,", pocet, GR, CP);

	  tr = floor (pocet * CP / 100.0);
	  notr = NE - tr;
	  if (verbose && CP != 100 && ps == 1)
	    printf ("i.e. %d elements\n", tr);

	  if (CP != 100)
	    for (k = 0; k < tr; k++)
	      {
		Extr[k] = k;
		for (j = 0; j < NE; j++)
		  if (Group[j] == GR)
		    {
		      if (EX == 0 && Val[j] > Val[Extr[k]])
			Extr[k] = j;
		      if (EX == 1 && Val[j] < Val[Extr[k]])
			Extr[k] = j;
		    }
		if (TO >= 0)
		  Group[Extr[k]] = TO;
		else
		  for (i = 0; i <= NS; i++)
		    if (i != GR && gn[i] != 0)
		      {
			Group[Extr[k]] = i;
			break;
		      }
	      }
/*
   for(i=0;i<=NS;i++)gn[i]=0;
   for(k=0;k<NE;k++)gn[Group[k]]++;
   if(verbose){for(i=0;i<=NS;i++) printf("%d ",gn[i]);printf("\n");}
 */
	  for (i = 0; i <= NS; i++)
	    gn[i] = 0;

	  for (j = 0; j < NE; j++)
	    {
	      i = Group[j];
	      val = Val[j];
	      gval2[i] += val * val;
	      gval[i] += val;
	      gn[i]++;
	    }
	  if (xfl == 1)
	    {
	      for (i = 0; i <= NS; i++)
		if (gn[i] != 0)
		  {
//orig       S = S + gval2[i] - gval[i] * gval[i] / gn[i];
		    S = S + gval2[i] - (gval[i] / gn[i]) * (gval[i] / gn[i]);
//		    if (verbose)
//		      printf ("%d %d %f %f %f\n", NS, gn[i], gval[i], gval2[i], S);
		  }
	    }
	  else
	    /* xfl==2 */
	    {
	      i = 0;
	      while (gn[i] == 0 && i <= NS)
		i++;
	      k = i;
	      i++;
	      while (gn[i] == 0 && i <= NS)
		i++;
	      if (gn[i] != 0)
		S = abs (gval[i] / gn[i] - gval[k] / gn[k]);
	      else
		printf ("smeg off!\n");
	    }
	}
      else
	/* correlation test */
	{
	  val = 0;
	  for (i = 0; i < NE; i++)
	    val = val + (cs[i].v - sx) * (cs[perm[i]].v1 - sy);
	  S = S + val / sqrt (vx * vy);
	}
    }
  if (ps == 0)
    min = S;
  if (S < min)
    pmin++;
  if (S == min)
    peq++;
  if (S > min)
    pmax++;

  if (verbose || Verbose)
    {
      fprintf (foutput, "%g\t", S);
      if (Verbose)
	for (i = 0; i < NE; i++)
	  fprintf (foutput, "%d ", perm[i]);
      fprintf (foutput, "\n");
    }
}


void
nperm (unsigned k, unsigned n)
{
  unsigned i, j, l;

  j = 0;
  for (i = 0; i < n - k; i++)
    {
      while (R[j] == 0)
	j++;
      perm[k] = j;
      R[j] = 0;
      if (k < n - 1)
	nperm (k + 1, n);
      else
	{
	  test ();
	  ps = 1;
	  NP++;
	}
      R[j] = 1;
      j++;
    }
}

int
main (int argc, char **argv)
{
  int i, j, k, l, m, n, pl, pr, pn;
  int c, p1, p;
  FILE *f1;
  double val, val1, val2, S;
  int optch;
  unsigned deff;
  char *start, *matx, *buf;
  vals *pv1, *pv2;
  node *s2;
  char sflag, ap;
  dvo *dp, *dn;
  cvo *cp, *cn;

  pfl = xfl = ap = pn = n = pl = pr = mt = 0;
  p1 = p = '(';
  NP = 10000;
  pv1 = pv2 = NULL;

  verbose = Verbose = 0;
  sflag = 1;
  EX = 0;
  CP = 100;
  GR = 0;
  TO = -1;

  foutput = stdout;
  while ((optch = getopt (argc, argv, "T:P:G:E:M:n:o:apvsVcxXm:h")) != EOF)
    {
      switch (optch)
	{
	case 'P':
	  start = optarg;
	  if (!(sscanf (start, "%lu", &CP)) ||
	      (CP < 0) || (CP >= 100))
	    {
	      fprintf (stderr, "-P percent in <0,100)\n");
	      exit (1);
	    }
	  break;
	case 'G':
	  start = optarg;
	  if (!(sscanf (start, "%lu", &GR)) ||
	      (GR < 0))
	    {
	      fprintf (stderr, "-G group >=0\n");
	      exit (1);
	    }
	  break;
	case 'T':
	  start = optarg;
	  if (!(sscanf (start, "%lu", &TO)) ||
	      (TO < 0))
	    {
	      fprintf (stderr, "-T group >=0\n");
	      exit (1);
	    }
	  break;
	case 'E':
	  start = optarg;
	  if (!(sscanf (start, "%lu", &EX)) ||
	      !(EX == 0 || EX == 1))
	    {
	      fprintf (stderr, "-E min|max=0|1\n");
	      exit (1);
	    }
	  break;
	case 'n':
	  start = optarg;
	  if ((!sscanf (start, "%lu", &deff)) ||
	      (deff < 0))
	    {
	      fprintf (stderr, "-n >=0");
	      exit (1);
	    }
	  NP = (unsigned) deff;
	  break;
	case 'm':
	  start = optarg;
	  if ((!sscanf (start, "%lu", &deff)) ||
	      (deff < 0))
	    {
	      fprintf (stderr, "-m >=0");
	      exit (1);
	    }



#ifdef _PLAN9
	  srand (deff);
#else
	  srandom (deff);
#endif
	  break;
	case 'o':
	  if ((foutput = fopen (optarg, "w")) == NULL)
	    {
	      fprintf (stderr, "Cannot open %s file for writing\n", optarg);
	      exit (1);
	    }
	  break;
	case 'a':
	  ap = 1;
	  break;
	case 'p':
	  pfl = 1;
	  break;
	case 'c':
	  xfl = 3;
	  break;
	case 'X':
	  xfl = 2;
	  break;
	case 'x':
	  xfl = 1;
	  break;
	case 'M':
	  matx = optarg;
	  mt = 1;
	  break;
	case 'v':
	  verbose = 1;
	  break;
	case 's':
	  sflag = 0;
	  break;
	case 'V':
	  Verbose = 1;
	  break;
	case 'h':
	default:
	  fprintf (stderr, "Usage: %s [OPTION] [FILE]\n", argv[0]);
	  fprintf (stderr, "Version: %s\n\n", VERSION);
	  fprintf (stderr, "-h\t\thelp\n");
	  fprintf (stderr, "-v\t\tverbose, print partial results\n");
	  fprintf (stderr, "-V\t\tVerbose, print permutations too\n");
	  fprintf (stderr, "-s\t\tsupress output of final results\n");
	  fprintf (stderr, "-p\t\tshow progress in percents\n");
	  fprintf (stderr, "-m seed\t\tset seed for random generator, >=0\n");
	  fprintf (stderr, "-n nr\t\tset number of random permutations, >=0\n");
	  fprintf (stderr, "-a \t\tgenerate and test all permutations, overrides -n \n");
	  fprintf (stderr, "-o file\t\tredirect output to file\n");
	  fprintf (stderr, "-x \t\tvariance test (ANOVA), -M is ignored\n");
	  fprintf (stderr, "-X \t\tvariance test, -M is ignored\n");
	  fprintf (stderr, "\nIn case of variance tests the following three args can be given (all together)\n");
	  fprintf (stderr, "-G group\tcut {min|max} %% of group\n");
	  fprintf (stderr, "-P percent\tcut {min|max} %% of group\n");
	  fprintf (stderr, "-E {0|1}\tcut {min|max} %% of group\n\n");
	  fprintf (stderr, "And yet another argument can be set in case of variance tests\n");
	  fprintf (stderr, "-T group\tmove elements to given group\n\n");
	  fprintf (stderr, "-c \t\tcorrelation test, -M is ignored\n");
	  fprintf (stderr, "-M file\t\tread matrix file\n\n");
	  fprintf (stderr, "With no FILE read standard input.\n");
	  exit (1);
	}
    }

  f1 = stdin;
  if (optind < argc)
    {
      f1 = fopen (argv[optind], "r");
      if (f1 == NULL)
	{
	  fprintf (stderr, "Cannot open file %s for reading\n", argv[optind]);
	  exit (1);
	}
    }

  if (xfl == 0)
    {
      pv1 = pv2 = (vals *) malloc (sizeof (vals));
      pv1->next = pv2->next = NULL;

      while ((c = fgetc (f1)) != '\n' && c != EOF)
	{
	  switch (c)
	    {
	    case ' ':
	      break;
	    case '(':
	      pl++;
	      if (pn == 2)
		{
		  fprintf (stderr, "ERR: NR NR (\n");
		  exit (1);
		}
	      set_node ('B', pl - 1, n);
	      push (pl);
	      if (p == 'n')
		{
		  if (p1 == ')')
		    {
		      fprintf (stderr, "ERR: ) NR (\n");
		      exit (1);
		    }
		  set_node ('M', pl - 2, n);
		}
	      if (p == ')')
		set_node ('M', i - 2, n);
	      pn = 0;
	      p = '(';
	      break;
	    case ')':
	      pr++;
	      i = pull ();
	      if (i == -1)
		{
		  fprintf (stderr, "SYNTAX ERR\n");
		  exit (1);
		}
	      set_node ('E', i - 1, n - 1);
	      if (p == 'n')
		{
		  if (p1 == '(')
		    {
		      fprintf (stderr, "ERR: ( NR )\n");
		      exit (1);
		    }
		}
	      if (p == '(')
		{
		  fprintf (stderr, "ERR: ()\n");
		  exit (1);
		}
	      pn = 0;
	      p1 = p;
	      p = ')';
	      break;
	    case '+':
	    case '-':
	    case '.':
	    case '0':
	    case '1':
	    case '2':
	    case '3':
	    case '4':
	    case '5':
	    case '6':
	    case '7':
	    case '8':
	    case '9':
	      ungetc (c, f1);
	      fscanf (f1, "%lg", &val);
	      pv2->val = val;
	      pv2->next = (vals *) malloc (sizeof (vals));
	      pv2 = pv2->next;
	      pv2->next = NULL;
	      pn++;
	      if (pn == 3)
		{
		  fprintf (stderr, "ERR: NR NR NR\n");
		  exit (1);
		}
	      if (p == 'n')
		set_node ('M', pl - 1, n);
	      if (p == ')')
		set_node ('M', i - 2, n);
	      p1 = p;
	      p = 'n';
	      n++;
	      break;
	    default:
	      fprintf (stderr, "ERR: bad character %d\n", c);
	      exit (1);
	    }
	}
      NE = n;
      NN = n - 1;
      ends = (double *) malloc (sizeof (double) * NE);
      nodes = (nod *) malloc (sizeof (nod) * NN);

      i = j = 0;

      while (pv1->next != NULL)
	{
	  ends[i] = pv1->val;
	  i++;
	  pv2 = pv1;
	  pv1 = pv1->next;
	  free (pv2);
	}
      free (pv1);

      i = 0;
      while (s1->prev != NULL)
	s1 = s1->prev;
      while (s1->next != NULL)
	{
	  nodes[i].B = s1->B;
	  nodes[i].M = s1->M;
	  nodes[i].E = s1->E;
	  s2 = s1;
	  s1 = s1->next;
	  i++;
	  free (s2);
	}
      nodes[i].B = s1->B;
      nodes[i].M = s1->M;
      nodes[i].E = s1->E;
      free (s1);
    }
  else
    {

#define BUF	10000

      if (xfl == 1 || xfl == 2)	/* input for variance test */
	{
	  NS = 0;
	  n = 0;
	  ds = dp = (dvo *) malloc (sizeof (dvo));
	  buf = (char *) malloc (BUF * sizeof (char));
	  while ((buf = fgets (buf, BUF, f1)) != NULL)
	    {
	      while (*buf == ' ' || *buf == '\t')
		buf++;
	      if (*buf != '\n' && *buf != '#')
		{
		  sscanf (buf, "%lg %lu", &val, &m);
		  dp->val = val;
		  dp->group = m;
		  if (m < 0)
		    {
		      fprintf (stderr, "group must be >=0\n");
		      exit (1);
		    }
		  if (m > NS)
		    NS = m;
		  dn = (dvo *) malloc (sizeof (dvo));
		  dp->next = dn;
		  dp = dn;
		  n++;
		}
	    }
	  free (dp);
	  NE = n;
	  dp = ds;
	  ds = (dvo *) malloc (sizeof (dvo) * n);
	  for (i = 0; i < n; i++)
	    {
	      ds[i].val = dp->val;
	      ds[i].group = dp->group;
	      dn = dp;
	      dp = dp->next;
	      free (dn);
	    }
/*
   for (i = 0; i < n; i++)
   printf ("%g %d\n", ds[i].val, ds[i].group);

   exit (0);
 */
	}
      else
	/* input for the correlation test */
	{
	  n = 0;
	  cs = cp = (cvo *) malloc (sizeof (cvo));
	  buf = (char *) malloc (BUF * sizeof (char));
	  while ((buf = fgets (buf, BUF, f1)) != NULL)
	    {
	      while (*buf == ' ' || *buf == '\t')
		buf++;
	      if (*buf != '\n' && *buf != '#')
		{
		  sscanf (buf, "%lg %lg", &val, &val1);
		  cp->v = val;
		  cp->v1 = val1;
		  cn = (cvo *) malloc (sizeof (cvo));
		  cp->next = cn;
		  cp = cn;
		  n++;
		}
	    }
	  free (cp);
	  NE = n;
	  cp = cs;
	  cs = (cvo *) malloc (sizeof (cvo) * n);
	  for (i = 0; i < n; i++)
	    {
	      cs[i].v = cp->v;
	      cs[i].v1 = cp->v1;
	      cn = cp;
	      cp = cp->next;
	      free (cn);
	    }
/*   for (i = 0; i < n; i++)
   printf ("%g\t%g\n", cs[i].v, cs[i].v1);
   exit (0);
 */
	  vx = sx = vy = sy = 0;
	  for (i = 0; i < n; i++)
	    {
	      sx += cs[i].v;
	      sy += cs[i].v1;
	      vx += cs[i].v * cs[i].v;
	      vy += cs[i].v1 * cs[i].v1;
	    }
	  vx = vx - sx * sx / n;
	  vy = vy - sy * sy / n;
	  sx = sx / n;
	  sy = sy / n;
	}
    }
  fclose (f1);

  if ((mt == 1) && (xfl == 0))
    {
      f1 = fopen (matx, "r");
      mat = (double *) malloc (NE * NE * sizeof (double));
      buf = (char *) malloc (BUF * sizeof (char));
      i = 0;
      while ((buf = fgets (buf, BUF, f1)) != NULL)
	{
	  j = 0;
	  while (*buf != '\n')
	    {
	      while (*buf == ' ' || *buf == '\t')
		buf++;

	      if (!(isdigit (*buf) || *buf == '+' || *buf == '-' || *buf == '.'))
		goto Z1;
	      val = strtod (buf, &start);
	      mat[i * NE + j] = val;
	      j++;
	      buf = start;
	    }
	Z1:;
	  if (j != 0)
	    i++;
	}
      fclose (f1);
    }


  perm = (unsigned *) malloc (sizeof (unsigned) * NE);
  R = (unsigned *) malloc (sizeof (unsigned) * NE);

  for (i = 0; i < NE; i++)
    perm[i] = i;

  peq = pmax = pmin = 0;
  if (ap == 0)
    {
      NPERM = NP;
      for (ps = 0; ps <= NP; ps++)
	{
	  for (i = 0; i < NE; i++)
	    R[i] = 1;
	  test ();
	  for (i = 0; i < NE; i++)
	    {
#ifdef _PLAN9
	      l = (int) (((double) NE - i) * rand() / (RAND_MAX + 1.0));
#else
	      l = (int) (((double) NE - i) * random() / (RAND_MAX + 1.0));
#endif
	      m = 0;
	      for (j = 0; j < NE; j++)
		{
		  if (R[j] == 1)
		    {
		      if (m == l)
			{
			  R[j] = 0;
			  k = j;
			  goto X1;
			}
		      m++;
		    }
		}
	    X1:;
	      perm[i] = k;
	    }
	}
    }
  else
    {
      NP = 0;
      ps = 0;
      if (NE > 12)
	{
	  fprintf (stderr, "Too many permutations... (length >12)\n");
	  exit (1);
	}
      for (i = 1; i <= NE; i++)
	NPERM *= i;
//      printf ("NPERM %u NE %u\n", NPERM,NE);
      for (i = 0; i < NE; i++)
	R[i] = 1;
      nperm (0, NE);
      NP--;
    }

  if (sflag)
    {
      fprintf (foutput, "\nResult:\t%g\n", min);
      if (NP > 0)
	{
	  fprintf (foutput, "Less:\t\t%d\t%.3g%%\n", pmin, (float) pmin / NP * 100);
	  fprintf (foutput, "Equal:\t\t%d\t%.3g%%\n", peq - 1, (float) (peq - 1) / NP * 100);
	  fprintf (foutput, "Greater:\t%d\t%.3g%%\n", pmax, (float) pmax / NP * 100);
	}
    }
  fclose (foutput);
}
